from ..logging import logg, settings
from .transition_matrix import transition_matrix
from .utils import scale
from scipy.sparse import linalg
import numpy as np


def eigs(T, k=10, eps=1e-3):                            # cred to M. Lange
    eigval, eigvec = linalg.eigs(T.T, k=k, which='LR')  # find k eigs with largest real part
    p = np.argsort(eigval)[::-1]                        # sort in descending order of eigenvalues
    eigval = eigval.real[p]
    eigvec = np.absolute(eigvec.real[:, p])

    eigvec = eigvec[:, (eigval >= 1 - eps)]             # select eigenvectors with eigenvalue of 1
    upper_bound = np.percentile(eigvec.flatten(), 99)   # clip to 99% quantile
    eigvec = np.clip(eigvec, 0, upper_bound)
    return eigval, eigvec


def terminal_states(adata, vkey='velocity', basis=None, diffuse_prob=0, eps=1e-3, copy=False):
    logg.info('computing root cells', r=True, end=' ')
    T = transition_matrix(adata, vkey=vkey, basis=basis, diffuse_prob=diffuse_prob, backward=True)
    _, eigvec = eigs(T, eps=eps)
    adata.obs['root'] = scale(eigvec.sum(1))
    logg.info('using ' + str(eigvec.shape[1]) + ' eigenvectors with eigenvalue 1.')

    logg.info('computing end points', end=' ')
    T = transition_matrix(adata, vkey=vkey, basis=basis)
    _, eigvec = eigs(T, eps=eps)
    adata.obs['end'] = scale(eigvec.sum(1))
    logg.info('using ' + str(eigvec.shape[1]) + ' eigenvectors with eigenvalue 1.')

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added to `.obs`\n'
        '    \'root\', root cells of Markov diffusion process\n'
        '    \'end\', end points of Markov diffusion process')
    return adata if copy else None



from ..logging import logg, settings
from ..preprocessing.moments import moments, second_order_moments
from .solver import solve_cov, solve2_inv, solve2_mle
from .utils import prod_sum_obs
import numpy as np
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)


def velocity(adata, vkey='velocity', mode='stochastic', fit_offset=False, fit_offset2=False, filter_genes=False, copy=False):
    """Estimates velocities in a gene-specific manner

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Name under which to refer to the computed velocities for `velocity_graph` and `velocity_embedding`.
    mode: `'deterministic'`, `'stochastic'` or `'bayes'` (default: `'stochastic'`)
        Whether to run the estimation using the deterministic or stochastic model of transcriptional dynamics.
        `'bayes'` solves the stochastic model and accounts for heteroscedasticity, but is slower than `'stochastic'`.
    fit_offset: `bool` (default: `False`)
        Whether to fit with offset for first order moment dynamics.
    fit_offset2: `bool`, (default: `False`)
        Whether to fit with offset for second order moment dynamics.
    filter_genes: `bool` (default: `True`)
        Whether to remove genes that are not used for further velocity analysis.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to `adata`.

    Returns
    -------
    Returns or updates `adata` with the attributes
    velocity: `.layers`
        velocity vectors for each individual cell
    variance_velocity: `.layers`
        velocity vectors for the cell variances
    velocity_offset, velocity_beta, velocity_gamma, velocity_r2: `.var`
        parameters
    """
    if 'Ms' not in adata.layers.keys(): moments(adata)
    Ms, Mu = adata.layers['Ms'], adata.layers['Mu']

    logg.info('computing velocities', r=True)

    beta = np.ones(adata.n_vars, dtype="float32")  # estimate all rates in units of the splicing rate
    offset, gamma = solve_cov(Ms, Mu, fit_offset)

    stochastic = any([mode in item for item in ['stochastic', 'bayes', 'alpha']])
    if stochastic:
        Mss, Mus = second_order_moments(adata)
        offset2, gamma2 = solve_cov(2 * Mss - Ms, 2 * Mus + Mu, fit_offset2)

        # initialize covariance matrix
        res_std = (Mu - gamma[None, :] * Ms - offset[None, :]).std(0)
        res2_std = (2 * Mus + Mu - gamma2[None, :] * (2 * Mss - Ms) - offset2[None, :]).std(0)

        # solve multiple regression
        offset, offset2, gamma = solve2_mle(Ms, Mu, Mus, Mss, fit_offset, fit_offset2) if mode == 'bayes' \
            else solve2_inv(Ms, Mu, 2 * Mss - Ms, 2 * Mus + Mu, res_std, res2_std, fit_offset, fit_offset2)

        adata.layers['variance_velocity'] = beta[None, :] * (2 * Mus - 2 * Ms * Mu + Mu) - \
                                    gamma[None, :] * (2 * Mss - 2 * Ms ** 2 - Ms) - \
                                    offset2[None, :] + 2 * offset[None, :] * Ms

        # if mode == 'alpha':
        #     Muu = second_order_moments_u(adata)
        #     offset2u, alpha = solve_cov(np.ones(Mu.shape) + 2 * Mu, 2 * Muu - Mu)
        #     pars.extend([offset2u, alpha])
        #     pars_str.extend(['_offset2u', '_alpha'])

    res, tot = Mu - gamma[None, :] * Ms - offset[None, :], Mu - Mu.mean(0)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        r2 = np.ones(adata.n_vars) - prod_sum_obs(res, res) / prod_sum_obs(tot, tot)

    gamma[np.isnan(gamma)], r2[np.isnan(r2)], res[np.isnan(res)] = 0, 0, 0

    pars = [offset, offset2, beta, gamma, r2] if stochastic else [offset, beta, gamma, r2]
    pars_str = ['_offset', '_offset2', '_beta', '_gamma', '_r2'] if stochastic else ['_offset', '_beta', '_gamma', '_r2']
    for i, par in enumerate(pars_str): adata.var[vkey + par] = pars[i]

    adata.layers[vkey] = Mu - gamma[None, :] * Ms - offset[None, :]

    adata.var['velocity_genes'] = (r2 > .01) & (gamma > .01)
    if filter_genes: adata._inplace_subset_var(adata.var['velocity_genes'])

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added to `.layers`\n'
        '    \'' + vkey + '\', velocity vectors for each individual cell')

    return adata if copy else None



import os
import numpy as np
import pandas as pd
from matplotlib import rcParams
import matplotlib.pyplot as pl
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from scanpy.plotting.utils import savefig_or_show
from ..logging import logg, settings


zeileis_26 = [
    "#023fa5", "#7d87b9", "#bec1d4", "#d6bcc0", "#bb7784", "#8e063b", "#4a6fe3",
    "#8595e1", "#b5bbe3", "#e6afb9", "#e07b91", "#d33f6a", "#11c638", "#8dd593",
    "#c6dec7", "#ead3c6", "#f0b98d", "#ef9708", "#0fcfc0", "#9cded6", "#d5eae7",
    "#f3e1eb", "#f6c4e1", "#f79cd4", '#7f7f7f', "#c7c7c7", "#1CE6FF", "#336600"]


def get_colors(adata, basis):
    clusters_colors = adata.uns[basis + '_colors'] if basis+'_colors' in adata.uns.keys() else zeileis_26
    cluster_ix = adata.obs[basis].cat.codes
    return np.array([clusters_colors[cluster_ix[i]] for i in range(adata.n_obs)])


def bound(c, perc=None):
    lb, ub = np.percentile(c, perc)
    return np.clip(c, lb, ub)


def interpret_colorkey(adata, c=None, layer=None, perc=None):
    if (c == 'louvain' or c == 'clusters') and adata.obs[c].dtype.name != 'category':
        adata.obs[c] = pd.Categorical(adata.obs[c])
    if isinstance(c, str):
        if c in adata.obs.keys():  # color by observation key
            c = get_colors(adata, basis=c) if adata.obs[c].dtype.name == 'category' else adata.obs[c]
        elif c in adata.var_names:  # color by var in specific layer
            c = adata[:, c].layers[layer] if layer in adata.layers.keys() else adata[:, c].X
        c = c if perc is None else bound(c, perc=perc)
    elif c is None:  # color by cluster or louvain or grey if no color is specified
        c = get_colors(adata, 'clusters') if 'clusters' in adata.obs.keys() \
            else get_colors(adata, 'louvain') if 'louvain' in adata.obs.keys() else 'grey'
    elif isinstance(c, np.ndarray) and len(c.flatten()) == len(adata.obs):  # continuous coloring
        c = c.flatten() if perc is None else bound(c.flatten(), perc=perc)
    else: raise ValueError('color key is invalid! pass valid observation annotation or a gene name')
    return c


def get_components(components=None):
    if components is None: components = '1,2'
    if isinstance(components, str): components = components.split(',')
    return np.array(components).astype(int) - 1


def quiver_autoscale(X, Y, U, V):
    Q = pl.quiver(X, Y, U, V, angles='xy', scale_units='xy', scale=None)
    Q._init()
    pl.clf()
    return Q.scale


def plot_colorbar(ax, orientation='vertical'):
    cb = pl.colorbar(orientation=orientation, cax=inset_axes(ax, width="2%", height="30%", loc=4, borderpad=0))
    cb.locator = (MaxNLocator(nbins=3))
    cb.update_ticks()


def savefig(writekey, show=False, dpi=None, save=None):
    """Save current figure to file.
    """
    savefig_or_show('velocity_' + writekey + '_' if writekey != '' else 'velocity_', dpi=dpi, save=save, show=show)


# def phase(adata, var=None, x=None, y=None, color='louvain', fits='all', xlabel='spliced', ylabel='unspliced',
#           fontsize=None, show=True, ax=None, **kwargs):
#     if isinstance(var, str) and (var in adata.var_names):
#         if (x is None) or (y is None):
#             ix = np.where(adata.var_names == var)[0][0]
#             x, y = adata.layers['Ms'][:, ix], adata.layers['Mu'][:, ix]
#     else:
#         ValueError('var not found in adata.var_names.')
#
#     ax = scatter(adata, x=x, y=y, color=color, frameon=True, title=var, xlabel=xlabel, ylabel=ylabel, ax=ax, **kwargs)
#
#     xnew = np.linspace(0, x.max() * 1.02)
#     fits = adata.layers.keys() if fits == 'all' else fits
#     fits = [fit for fit in fits if 'velocity' in fit]
#     for fit in fits:
#         linestyle = '--' if 'stochastic' in fit else '-'
#         pl.plot(xnew, adata.var[fit+'_gamma'][ix] / adata.var[fit+'_beta'][ix] * xnew
#                 + adata.var[fit+'_offset'][ix] / adata.var[fit+'_beta'][ix], c='k', linestyle=linestyle)
#
#     if show: pl.show()
#     else: return ax


import numpy as np


def show_proportions(adata, copy=False):
    """Fraction of spliced/unspliced/ambiguous abundances

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to adata.

    Returns
    -------
    Prints the fractions of abundances.
    """
    layers_keys = [key for key in ['spliced', 'unspliced', 'ambiguous'] if key in adata.layers.keys()]
    tot_mol_cell_layers = [adata.layers[key].sum(1) for key in layers_keys]

    mean_abundances = np.round(
        [np.mean(tot_mol_cell / np.sum(tot_mol_cell_layers, 0)) for tot_mol_cell in tot_mol_cell_layers], 2)

    print('abundance of ' + str(layers_keys) + ': ' + str(mean_abundances))

    return adata if copy else None


def cleanup(adata, clean='layers', keep={'spliced', 'unspliced'}, copy=False):
    """Deletes attributes not needed.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    cleanup: `str` or list of `str` (default: `layers`)
        Which attributes to consider for freeing memory.
    keep: `str` or list of `str` (default: `['spliced', unspliced']`)
        Which attributes to keep.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to adata.

    Returns
    -------
    Returns or updates `adata` with selection of attributes kept.
    """

    if any(['obs' in clean, 'all' in clean]):
        for key in list(adata.obs.keys()):
            if key not in keep: del adata.obs[key]

    if any(['var' in clean, 'all' in clean]):
        for key in list(adata.var.keys()):
            if key not in keep: del adata.var[key]

    if any(['obsm' in clean, 'all' in clean]):
        for key in list(adata.obsm.keys()):
            if key not in keep: del adata.obsm[key]

    if any(['varm' in clean, 'all' in clean]):
        for key in list(adata.varm.keys()):
            if key not in keep: del adata.varm[key]

    if any(['uns' in clean, 'all' in clean]):
        for key in list(adata.uns.keys()):
            if key not in keep: del adata.uns[key]

    if any(['layers' in clean, 'all' in clean]):
        for key in list(adata.layers.keys()):  # remove layers that are not needed
            if key not in keep: del adata.layers[key]

    return adata if copy else None


def filter_and_normalize(adata, min_counts=10, n_top_genes=4000, log=True, copy=False):
    """Filtering, normalization and log transform

    Expects non-logarithmized data. If using logarithmized data, pass `log=False`.

    Runs the following steps

    .. code:: python

        sc.pp.filter_genes(adata, min_counts=10)
        sc.pp.normalize_per_cell(adata)
        sc.pp.filter_genes_dispersion(adata, n_top_genes=10000)
        sc.pp.normalize_per_cell(adata)
        if log: sc.pp.log1p(adata)


    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    min_counts: `int` (default: 10)
        Minimum number of gene counts per cell.
    n_top_genes: `int` (default: 10000)
        Number of genes to keep.
    log: `bool` (default: `True`)
        Take logarithm.
    copy: `bool` (default: `False`)
        Return a copy of `adata` instead of updating it.

    Returns
    -------
    Returns or updates `adata` depending on `copy`.
    """
    from scanpy.api.pp import filter_genes, filter_genes_dispersion, normalize_per_cell, log1p
    filter_genes(adata, min_counts=min_counts)
    if n_top_genes is None or n_top_genes < adata.shape[1]:
        normalize_per_cell(adata)
        filter_genes_dispersion(adata, n_top_genes=n_top_genes)
    normalize_per_cell(adata)
    if log: log1p(adata)
    return adata if copy else None


def recipe_velocity(adata, min_counts=10, n_top_genes=None, n_pcs=30, n_neighbors=30, log=True, copy=False):
    """Runs pp.filter_and_normalize() and pp.moments()
    """
    from .moments import moments
    filter_and_normalize(adata, min_counts=min_counts, n_top_genes=n_top_genes, log=log)
    moments(adata, n_neighbors=n_neighbors, n_pcs=n_pcs)
    return adata if copy else None



from ..logging import logg, settings
from .utils import norm
from .transition_matrix import transition_matrix
import numpy as np
import warnings


def velocity_embedding(adata, basis='tsne', vkey='velocity', scale=10, retain_scale=False, copy=False):
    """Computes the single cell velocities in the embedding

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    basis: `str` (default: `'tsne'`)
        Which embedding to use.
    vkey: `str` (default: `'velocity'`)
        Name of velocity estimates to be used.
    scale: `int` (default: 10)
        Scale parameter of gaussian kernel for transition matrix.
    retain_scale: `bool` (default: `False`)
        Whether to retain scale from high dimensional space in embedding.

    Returns
    -------
    Returns or updates `adata` with the attributes
    velocity_umap: `.obsm`
        coordinates of velocity projection on embedding
    """
    T = transition_matrix(adata, vkey=vkey, basis=basis, scale=scale)

    logg.info('computing velocity embedding', r=True)

    if 'X_' + basis not in adata.obsm_keys():
        raise ValueError(
            'You need to run `tl.{}` first to compute embedded velocity vectors.').format(basis)

    X_emb = adata.obsm['X_' + basis][:, :2]
    V_emb = np.zeros((adata.n_obs, 2))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if adata.n_obs < 8192:
            TA = T.A
            for i in range(adata.n_obs):
                indices = T[i].indices
                diff = X_emb[indices] - X_emb[i, None]
                diff /= norm(diff)[:, None]
                diff[np.isnan(diff)] = 0  # zero diff in a steady-state
                V_emb[i] = TA[i, indices].dot(diff) - diff.sum(0) / len(indices)
        else:
            for i in range(adata.n_obs):
                indices = T[i].indices
                diff = X_emb[indices] - X_emb[i, None]
                diff /= norm(diff)[:, None]
                diff[np.isnan(diff)] = 0  # zero diff in a steady-state
                V_emb[i] = T[i].data.dot(diff) - diff.sum(0) / len(indices)

    if retain_scale:
        delta = T.dot(adata.X) - adata.X
        cos_proj = (adata.obsm[vkey] * delta).sum(1) / norm(delta)
        V_emb *= np.clip(cos_proj, 0, 1)

    vkey += '_' + basis
    adata.obsm[vkey] = V_emb

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added to `.obsm`\n'
        '    \'' + vkey + '\', embedded velocity vectors')

    return adata if copy else None




from ..logging import logg, settings
from .utils import norm
from .velocity import velocity
from scanpy.api import Neighbors
from scipy.sparse import coo_matrix
import numpy as np
import warnings


def get_indices(dist):
    n_neighbors = (dist > 0).sum(1).min()
    rows_idx = np.where((dist > 0).sum(1) > n_neighbors)[0]

    for row_idx in rows_idx:
        col_idx = dist[row_idx].indices[n_neighbors:]
        dist[row_idx, col_idx] = 0

    dist.eliminate_zeros()

    indices = dist.indices.reshape((-1, n_neighbors))
    return indices, dist


def get_iterative_indices(indices, index, n_recurse_neighbors):
    return indices[get_iterative_indices(indices, index, n_recurse_neighbors-1)] \
        if n_recurse_neighbors > 1 else indices[index]


class Cosines:
    def __init__(self, X, V, indices, n_recurse_neighbors, sqrt_transform=False):
        self.X = X.copy()
        self.indices = indices
        self.n_recurse_neighbors = n_recurse_neighbors
        self.sqrt_transform = sqrt_transform

        self.V = V.copy()
        if sqrt_transform: self.V = np.sqrt(np.abs(self.V)) * np.sign(self.V)
        self.V = self.V - self.V.mean(1)[:, None]
        self.V_norm = norm(self.V)

    def compute(self, ixs):
        vals, rows, cols, async_id = [], [], [], []

        if self.sqrt_transform:
            for i in ixs:
                knn_ixs = np.unique(get_iterative_indices(self.indices, i, self.n_recurse_neighbors))
                dX = self.X[knn_ixs] - self.X[i, None]
                dX = np.sqrt(np.abs(dX)) * np.sign(dX)
                dX -= dX.mean(1)[:, None]

                vals.extend(np.einsum('ij, j', dX, self.V[i]) / (norm(dX) * self.V_norm[i])[None, :])
                rows.extend(np.ones(len(knn_ixs)) * i)
                cols.extend(knn_ixs)

        else:
            for i in ixs:
                knn_ixs = np.unique(get_iterative_indices(self.indices, i, self.n_recurse_neighbors))
                dX = self.X[knn_ixs] - self.X[i, None]
                dX -= dX.mean(1)[:, None]

                vals.extend(np.einsum('ij, j', dX, self.V[i]) / (norm(dX) * self.V_norm[i])[None, :])
                rows.extend(np.ones(len(knn_ixs)) * i)
                cols.extend(knn_ixs)
        async_id = int(ixs[0] / len(ixs))  # keep track of order

        vals = np.hstack(vals)
        vals[np.isnan(vals)] = 0
        vals = np.clip(vals, 1e-10, 1)

        return [async_id, vals, rows, cols]


def velocity_graph(adata, vkey='velocity', n_recurse_neighbors=2, n_neighbors=None, sqrt_transform=False, copy=False):
    """Computes a velocity graph based on cosine similarities.

    The cosine similarities are computed between velocities and potential cell state transitions.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Name of velocity estimates to be used.
    n_neighbors: `int` or `None` (default: None)
        Use fixed number of neighbors or do recursive neighbor search (if `None`).
    n_recurse_neighbors: `int` (default: 2)
        Number of recursions to be done for neighbors search.
    sqrt_transform: `bool` (default: `False`)
        Whether to variance-transform the cell states and velocities before computing cosine similarities.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to adata.

    Returns
    -------
    Returns or updates `adata` with the attributes
    velocity_graph: `.uns`
        sparse matrix with transition probabilities
    """
    if vkey not in adata.layers.keys(): velocity(adata)

    logg.info('computing velocity graph', r=True)

    if n_neighbors is not None:
        keys = [key for key in ['X_pca', 'X_tsne', 'X_umap'] if key in adata.obsm.keys()]
        neighs = Neighbors(adata)
        neighs.compute_neighbors(n_neighbors=n_neighbors, use_rep=keys[-1], n_pcs=10)
        indices = get_indices(dist=neighs.distances)[0]
        n_recurse_neighbors = 1
    else:
        indices = get_indices(dist=adata.uns['neighbors']['distances'])[0]

    cos = Cosines(adata.layers['Ms'][:, adata.var['velocity_genes']],
                  adata.layers[vkey][:, adata.var['velocity_genes']], indices, n_recurse_neighbors, sqrt_transform)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if True:
            _, vals, rows, cols = cos.compute(range(adata.n_obs))  # list(map(cos.compute, range(adata.n_obs)))
        else:
            from multiprocessing import Pool
            pool = Pool(n_jobs)
            ranges = np.array_split(range(adata.n_obs), n_jobs)
            result = pool.map_async(cos.compute, ranges).get()  # async will probably destroy order
            result.sort(key=lambda tup: tup[0])  # restore order
            vals, rows, cols = [], [], []
            for r in result:
                vals.extend(r[1])
                rows.extend(r[2])
                cols.extend(r[3])

    graph = coo_matrix((vals, (rows, cols)), shape=(adata.n_obs, adata.n_obs))
    graph.eliminate_zeros()

    adata.uns[vkey+'_graph'] = graph.tocsr()

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added to `.uns`\n'
        '    \'' + vkey + '_graph\', sparse matrix with cosine correlations')

    return adata if copy else None



from .velocity import velocity
from .velocity_graph import velocity_graph
from .transition_matrix import transition_matrix
from .velocity_embedding import velocity_embedding
from .velocity_score import score_smoothness, score_transition, score_robustness
from .terminal_states import eigs, terminal_states



from scipy.sparse import csr_matrix
import numpy as np
import warnings


def prod_sum_obs(A, B):
    """dot product and sum over axis 0 (obs) equivalent to np.sum(A * B, 0)
    """
    return np.einsum('ij, ij -> j', A, B)


def prod_sum_var(A, B):
    """dot product and sum over axis 1 (var) equivalent to np.sum(A * B, 1)
    """
    return np.einsum('ij, ij -> i', A, B)


def norm(A):
    """computes the L2-norm along axis 1 (e.g. genes or embedding dimensions) equivalent to np.linalg.norm(A, axis=1)
    """
    return np.sqrt(np.einsum('ij, ij -> i', A, A))


def normalize_sparse(X):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X = X.multiply(csr_matrix(1. / X.sum(1)))
    return X


def normalize_dense(X):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X = X / X.sum(1)
    return X


def scale(X, min=0, max=1):
    X = X - X.min() + min
    X = X / X.max() * max
    return X



from .utils import interpret_colorkey, get_components, plot_colorbar, savefig
import matplotlib.pyplot as pl
import scanpy.api.pl as scpl
from matplotlib.ticker import MaxNLocator


def scatter(adata, x=None, y=None, basis='umap', layer=None, color=None, xlabel=None, ylabel=None, color_map=None,
            perc=None, size=5, alpha=1, fontsize=None, colorbar=False, groups=None, use_raw=None, sort_order=True,
            legend_loc='none', legend_fontsize=None, legend_fontweight=None, projection='2d', palette=None,
            right_margin=None, left_margin=None, components=None, frameon=False, title=None, show=True, figsize=(7,5),
            dpi=80, save=None, ax=None, zorder=0, **kwargs):
    """Scatter plot along observations or variables axes.
    Color the plot using annotations of observations (`.obs`), variables (`.var`) or expression of genes (`.var_names`).

    Arguments
    ---------
    adata: `AnnData`
        Annotated data matrix.

    basis: `str` (default='tsne')
        plots embedding obsm['X_' + basis]

    x: `str`, `np.ndarray` or `None` (default: `None`)
        x coordinate

    y: `str`, `np.ndarray` or `None` (default: `None`)
        y coordinate

    color : `str` or `None` (default: `None`)
        Key for annotations of observations/cells or variables/genes

    Returns
    -------
        If `show==False` a `matplotlib.Axis`
    """
    colors = color if isinstance(color, (list, tuple)) else [color]
    layers = layer if isinstance(layer, (list, tuple)) else [layer]

    if len(colors) > 1:
        for i, gs in enumerate(pl.GridSpec(1, len(colors), pl.figure(None, (figsize[0]*len(colors), figsize[1]), dpi=dpi))):
            scatter(adata, basis=basis, layer=layer, color=color[i], xlabel=xlabel, ylabel=ylabel, color_map=color_map,
                    perc=perc, size=size, alpha=alpha, fontsize=fontsize, frameon=frameon, title=title, show=False,
                    colorbar=colorbar, components=components, figsize=figsize, dpi=dpi, save=None, ax=pl.subplot(gs),
                    use_raw=use_raw, sort_order=sort_order, groups=groups, projection=projection,
                    legend_loc=legend_loc, legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight,
                    palette=palette, right_margin=right_margin, left_margin=left_margin, **kwargs)
        if isinstance(save, str): savefig('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if show: pl.show()
        else: return ax

    elif len(layers) > 1:
        for i, gs in enumerate(pl.GridSpec(1, len(layers), pl.figure(None, (figsize[0] * len(layers), figsize[1]), dpi=dpi))):
            scatter(adata, basis=basis, layer=layers[i], color=color, xlabel=xlabel, ylabel=ylabel, color_map=color_map,
                    perc=perc, size=size, alpha=alpha, fontsize=fontsize, frameon=frameon, title=title, show=False,
                    colorbar=colorbar, components=components, figsize=figsize, dpi=dpi, save=None, ax=pl.subplot(gs),
                    use_raw=use_raw, sort_order=sort_order, groups=groups, projection=projection,
                    legend_loc=legend_loc, legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight,
                    palette=palette, right_margin=right_margin, left_margin=left_margin, **kwargs)
        if isinstance(save, str): savefig('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if show: pl.show()
        else: return ax

    else:
        if ax is None: ax = pl.figure(None, figsize, dpi=dpi).gca()
        if color_map is None: color_map = 'viridis_r' if (color == 'root' or color == 'end') else 'RdBu_r'
        is_embedding = (x is None) | (y is None)

        if color in adata.obs and adata.obs[color].dtype.name == 'category' and is_embedding:
            ax = scpl.scatter(adata, color=color, use_raw=use_raw, sort_order=sort_order, alpha=alpha, basis=basis,
                              groups=groups, components=components, projection=projection, legend_loc=legend_loc,
                              legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight, color_map=color_map,
                              palette=palette, right_margin=right_margin, left_margin=left_margin,
                              size=size, title=title, frameon=frameon, show=False, save=None, ax=ax, **kwargs)
        else:
            if is_embedding:
                X_emb = adata.obsm['X_' + basis][:, get_components(components)]
                x, y = X_emb[:, 0], X_emb[:, 1]
                ax.tick_params(which='both', bottom=False, left=False, labelbottom=False, labelleft=False)
            else:
                ax.xaxis.set_major_locator(MaxNLocator(nbins=3))
                ax.yaxis.set_major_locator(MaxNLocator(nbins=3))
                labelsize = int(fontsize * .75) if fontsize is not None else None
                ax.tick_params(axis='both', which='major', labelsize=labelsize)
            c = interpret_colorkey(adata, color, layer, perc)
            pl.scatter(x, y, c=c, cmap=color_map, s=size, alpha=alpha, zorder=zorder, **kwargs)

            if isinstance(xlabel, str) and isinstance(ylabel, str):
                pl.xlabel(xlabel, fontsize=fontsize)
                pl.ylabel(ylabel, fontsize=fontsize)
            elif basis is not None:
                component_name = ('DC' if basis == 'diffmap' else 'tSNE' if basis == 'tsne' else 'UMAP' if basis == 'umap'
                else 'PC' if basis == 'pca' else basis.replace('draw_graph_', '').upper() if 'draw_graph' in basis else basis)
                pl.xlabel(component_name + '1')
                pl.ylabel(component_name + '2')

            if isinstance(title, str): pl.title(title, fontsize=fontsize)
            elif isinstance(layer, str) and isinstance(color, str): pl.title(color + ' ' + layer, fontsize=fontsize)
            elif isinstance(color, str): pl.title(color, fontsize=fontsize)

            if not frameon: pl.axis('off')
            if colorbar: plot_colorbar(ax)

        if isinstance(save, str): savefig('' if basis is None else basis, dpi=dpi, save=save, show=show)

        if show: pl.show()
        else: return ax


from ..tools.velocity_embedding import velocity_embedding as tl_velocity_embedding
from .utils import interpret_colorkey, quiver_autoscale, get_components, savefig
from .scatter import scatter
from matplotlib.colors import is_color_like
import matplotlib.pyplot as pl
import numpy as np


def velocity_embedding(adata, basis='umap', vbasis='velocity', layer=None, density=1, scale=1, autoscale=True,
                       perc=None, color=None, use_raw=True, sort_order=True, alpha=.2, groups=None, components=None,
                       projection='2d', legend_loc='none', legend_fontsize=None, legend_fontweight=None,
                       color_map=None, palette=None, frameon=False, right_margin=None, left_margin=None,
                       size=1, title=None, show=True, figsize=(14,10), dpi=150, save=None, ax=None,
                       xlabel=None, ylabel=None, colorbar=False, fontsize=None, **kwargs):
    """Scatter plot with velocities along `.obs` or `.var` axes.
    Color the plot using annotations of observations (`.obs`), variables (`.var`) or expression of genes (`.var_names`).

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    basis: `str` (default: `'umap'`)
        Key for embedding coordinates.
    vbasis: `str` (default: `'velocity'`)
        Key for velocity embedding coordinates.
    color : `str` or `None` (default: `None`)
        Key for annotations of observations/cells or variables/genes.

    Returns
    -------
        `matplotlib.Axis` if `show==False`
    """
    vkey = vbasis + '_' + basis
    if vkey not in adata.obsm_keys(): tl_velocity_embedding(adata, basis=basis, vkey=vkey)

    ix_choice = np.random.choice(adata.n_obs, size=int(density * adata.n_obs), replace=False)
    X = adata.obsm['X_' + basis][:, get_components(components)][ix_choice]
    V = adata.obsm[vkey][ix_choice]
    if autoscale: scale *= 2 * quiver_autoscale(X[:, 0], X[:, 1], V[:, 0], V[:, 1])

    colors = color if isinstance(color, (list, tuple)) else [color]
    layers = layer if isinstance(layer, (list, tuple)) else [layer]

    if len(colors) > 1:
        for i, gs in enumerate(pl.GridSpec(1, len(colors), pl.figure(None, (figsize[0] * len(colors), figsize[1]), dpi=dpi))):
            velocity_embedding(adata, basis=basis, vbasis=vbasis, layer=layer, density=density, scale=scale,
                               autoscale=False, perc=perc, color=colors[i], use_raw=use_raw, sort_order=sort_order,
                               alpha=alpha, groups=groups, components=components, projection=projection, legend_loc=legend_loc,
                               legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight,
                               color_map=color_map, palette=palette, frameon=frameon, right_margin=right_margin,
                               left_margin=left_margin, size=size, title=title, show=False, figsize=figsize, dpi=dpi,
                               save=None, ax=pl.subplot(gs), xlabel=xlabel, ylabel=ylabel, colorbar=colorbar,
                               fontsize=fontsize, **kwargs)
        if isinstance(save, str): savefig('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if show: pl.show()
        else: return ax

    elif len(layers) > 1:
        for i, gs in enumerate(pl.GridSpec(1, len(layers), pl.figure(None, (figsize[0] * len(layers), figsize[1]), dpi=dpi))):
            velocity_embedding(adata, basis=basis, vbasis=vbasis, layer=layers[i], density=density, scale=scale, autoscale=False,
                               perc=None, color=color, use_raw=use_raw, sort_order=sort_order, alpha=alpha,
                               groups=groups, components=components, projection=projection, legend_loc=legend_loc,
                               legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight,
                               color_map=color_map, palette=palette, frameon=frameon, right_margin=right_margin,
                               left_margin=left_margin, size=size, title=title, show=False, figsize=figsize, dpi=dpi,
                               save=None, ax=pl.subplot(gs), xlabel=xlabel, ylabel=ylabel, colorbar=colorbar,
                               fontsize=fontsize, **kwargs)
        if isinstance(save, str): savefig('' if basis is None else basis, dpi=dpi, save=save, show=show)
        if show: pl.show()
        else: return ax

    else:
        ix_choice = np.random.choice(adata.n_obs, size=int(density * adata.n_obs), replace=False)
        X = adata.obsm['X_' + basis][:, get_components(components)][ix_choice]
        V = adata.obsm[vkey][ix_choice]

        if color_map is None: color_map = 'viridis_r' if (color == 'root' or color == 'end') else 'RdBu_r'
        _kwargs = {"scale": scale, "cmap": color_map, "angles": 'xy', "scale_units": 'xy', "width": .0005,
                   "edgecolors": 'k', "headwidth": 9, "headlength": 10, "headaxislength": 6, "linewidth": .25}
        _kwargs.update(kwargs)

        if ax is None: ax = pl.figure(None, figsize, dpi=dpi).gca()

        C = interpret_colorkey(adata, color, layer, perc)[ix_choice]
        if is_color_like(C[0]): pl.quiver(X[:, 0], X[:, 1], V[:, 0], V[:, 1], color=C, zorder=1, **_kwargs)
        else: pl.quiver(X[:, 0], X[:, 1], V[:, 0], V[:, 1], C, zorder=1, **_kwargs)

        ax = scatter(adata, basis=basis, layer=layer, color=color, xlabel=xlabel, ylabel=ylabel, color_map=color_map,
                     perc=perc, size=size, alpha=alpha, fontsize=fontsize, frameon=frameon, title=title, show=False,
                     colorbar=colorbar, components=components, figsize=figsize, dpi=dpi, save=None, ax=ax, zorder=0,
                     use_raw=use_raw, sort_order=sort_order, groups=groups, projection=projection,
                     legend_loc=legend_loc, legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight,
                     palette=palette, right_margin=right_margin, left_margin=left_margin, ** kwargs)

        if isinstance(save, str): savefig('' if basis is None else basis, dpi=dpi, save=save, show=show)

        if show: pl.show()
        else: return ax



import scvelo as scv
import numpy as np


def test_einsum():
    from scvelo.tools.utils import prod_sum_obs, prod_sum_var, norm
    Ms, Mu = np.random.rand(5, 4), np.random.rand(5, 4)
    assert np.allclose(prod_sum_obs(Ms, Mu), np.sum(Ms * Mu, 0))
    assert np.allclose(prod_sum_var(Ms, Mu), np.sum(Ms * Mu, 1))
    assert np.allclose(norm(Ms), np.linalg.norm(Ms, axis=1))


def test_velocity_graph():
    adata = scv.datasets.toy_data(n_obs=500)
    scv.pp.recipe_velocity(adata, n_top_genes=300)
    scv.tl.velocity(adata)

    scv.tl.velocity_graph(adata)
    graph = adata.uns['velocity_graph']

    assert np.allclose((scv.tl.transition_matrix(adata) > 0).toarray(), (graph > 0).toarray())



from scipy.spatial.distance import pdist, squareform
from .utils import normalize_sparse
import numpy as np


def transition_matrix(adata, vkey='velocity', basis=None, backward=False, diffuse_prob=0, scale=10):
    """Computes transition probabilities by applying Gaussian kernel to cosine similarities x scale

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Name of velocity estimates to be used.
    basis: `str` or `None` (default: `None`)
        Restrict transition to embedding if specified
    backward: `bool` (default: `False`)
        Whether to use the transition matrix to push forward (`False`) or to pull backward (`True`)
    diffuse_prob:
        Relative weight to be given to diffusion kernel (Brownian motion)
    scale: `float` (default: 10)
        Scale parameter of gaussian kernel.

    Returns
    -------
    Returns sparse matrix with transition probabilities.
    """
    if vkey+'_graph' not in adata.uns:
        raise ValueError(
            'You need to run `tl.velocity_graph` first to compute cosine correlations.')

    T = np.expm1(adata.uns[vkey + '_graph'] * scale)

    # weight direct neighbors with 1 and indirect (recurse) neighbors with 0.5
    T = .5 * T + .5 * (adata.uns['neighbors']['distances'] > 0).multiply(T)

    if backward: T = T.T
    T = normalize_sparse(T)

    if 'X_' + str(basis) in adata.obsm.keys():
        dists_emb = (T > 0).multiply(squareform(pdist(adata.obsm['X_' + basis])))
        sigma = dists_emb.mean()
        diffusion_kernel = np.expm1(-.5 * dists_emb.multiply(dists_emb) / sigma**2)
        T = T.multiply(diffusion_kernel)  # combine velocity based kernel with diffusion based kernel
        if diffuse_prob > 0:  # add another diffusion kernel (Brownian motion - like)
            T = (1-diffuse_prob) * T + diffuse_prob * np.expm1(-.5 * dists_emb.multiply(dists_emb) / (sigma/2)**2)
        T = normalize_sparse(T)

    return T